#=======================================================================
#	$Id: LRU.pm,v 1.1 2006/06/07 12:25:37 pythontech Exp $
#	Least-recently-used hash-cache thingy.
#	Copyright (C) 2006  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	Cache of recently used lookup values.
#	When full, least recently used item is forgotten.
#
#	Touch existing item:
#		Set .seq to ++time (increases .seq)
#		Filter down
#	Add new:
#		If heap full:
#			Drop item[0]
#			Insert new at [0] with .seq = ++time
#			Filter down
#		Else:
#			Insert new at end with .seq = ++time
#	Delete:
#		Drop item[0]
#		Unless was only item:
#			Move last item to [0]
#			If cur[0].seq > deleted.seq
#				Filter down
#			Else:
#				Filter up
#=======================================================================
package PythonTech::LRU;
use strict;

# Have both a hash and a heap referring to these records
sub SEQ {0}			# Time of last access
sub POS {1}			# Position in heap (0..USED-1)
sub KEY {2}			# Hash key
sub VALUE {3}			# Hash value
# LRU is an array
sub SIZE {0}			# Max size of cache
sub USED {1}			# Entries in cache
sub HASH {2}			# Key => record
sub TIME {3}			# Access time counter
# Items 3 onwards form a heap: pos => [seq, key]
sub HEAP {4 + $_[0]}

sub new {
    my($class,%props) = @_;
    $props{'size'} ||= 20;
    my $self = [$props{'size'}, 0, {}, 0];
    bless $self, $class;
    return $self;
}

sub get {
    my($self, $key) = @_;
    my $rec = $self->[HASH]->{$key};
    return unless defined $rec;
    
    # Touch it
    $rec->[SEQ] = ++ $self->[TIME];
    $self->_filter_down($rec);
    return $rec->[VALUE];
}

sub put {
    my($self, $key,$value) = @_;
    my $rec = $self->[HASH]->{$key};
    if (defined $rec) {
	# Already present - update value and adjust seq
	$rec->[VALUE] = $value;
	$rec->[SEQ] = ++ $self->[TIME];
	$self->_filter_down($rec);
    } elsif ($self->[USED] < $self->[SIZE]) {
	# Just add at the end
	my $pos = $self->[USED] ++;
#	print STDERR "adding pos=$pos\n";
	my $rec = [++ $self->[TIME], $pos, $key, $value];
	$self->[HASH]->{$key} = $rec;
	$self->[HEAP($pos)] = $rec;
	# No need to filter - this is most recently used
    } else {
	# Need to replace the least recently used (pos=0)
	my $oldkey = $self->[HEAP(0)]->[KEY];
	delete $self->[HASH]->{$oldkey};
	my $rec = [++ $self->[TIME], 0, $key, $value];
	$self->[HASH]->{$key} =
	    $self->[HEAP(0)] =
		$rec;
	$self->_filter_down($rec);
    }
#    $self->_dump("done put $key");
    return $value;
}

sub delete {
    my($self, $key) = @_;
    my $oldrec = delete $self->[HASH]->{$key};
    if (defined $oldrec) {
	my $oldpos = $oldrec->[POS];
	my $used = -- $self->[USED];
	if ($used == 0) {
	    # Deleting the only record
	    # ASSERT: $oldpos == 0
	    $self->[HEAP($oldpos)] = undef;
	} else {
	    # Move last to vacant slot
	    my $rec = $self->[HEAP($used)];
	    $self->[HEAP($oldpos)] = $rec;
	    $rec->[POS] = $oldpos;
	    if ($rec->[SEQ] > $oldrec->[SEQ]) {
		$self->_filter_down($rec);
	    } else {
		$self->_filter_up($rec);
	    }
	}
    }
#    $self->_dump("delete $key");
    return $oldrec->[VALUE];
}

#-----------------------------------------------------------------------
#	The record at [i] may need to be moved down
#-----------------------------------------------------------------------
sub _filter_down {
    my($self, $irec) = @_;
#    $self->_dump("_filter_down $irec->[POS]");
    my $used = $self->[USED];
    # Get the record at position [i] - replace at end
    my $i = $irec->[POS];
    my $iseq = $irec->[SEQ];
    my($j,$k,$jrec,$krec,$jseq,$kseq);
    while ($j = 2*$i+1,
	   $j < $used) {
	$jrec = $self->[HEAP($j)];
	$jseq = $jrec->[SEQ];
	if (($k = $j+1) < $used &&
	    ($krec = $self->[HEAP($k)],
	     $kseq = $krec->[SEQ],
	     $kseq < $jseq)) {
	    # Swap with k if any
	    if ($iseq > $kseq) {
		# Swap i <=> k
#		print STDERR "Swap $i/$k\n";
		$self->[HEAP($i)] = $krec;
		$krec->[POS] = $i;
		$i = $k;
	    } else {
		last;
	    }
	} else {
	    if ($iseq > $jseq) {
		# Swap i <=> j
#		print STDERR "Swap $i/$j\n";
		$self->[HEAP($i)] = $jrec;
		$jrec->[POS] = $i;
		$i = $j;
	    } else {
		last;
	    }
	}
    }
    $self->[HEAP($i)] = $irec;
    $irec->[POS] = $i;
#    $self->_dump("After _filter_down");
}

#-----------------------------------------------------------------------
#	The record at [i] may need to be moved up
#-----------------------------------------------------------------------
sub _filter_up {
    my($self, $irec) = @_;
#    print STDERR "_filter_up $irec->[POS]\n";
    my $i = $irec->[POS];
    my($j);
    while ($i > 0) {
	my $j = int(($j-1) / 2);
	my $jrec = $self->[HEAP($j)];
	my $jseq = $jrec->[SEQ];
	last if $irec->[SEQ] > $jseq;		# >>> All in order
	# Move current occupant of [j] down to [i]
	$self->[HEAP($i)] = $jrec;
	$jrec->[POS] = $i;
	$i = $j;
    }
    $self->[HEAP($i)] = $irec;
    $irec->[POS] = $i;
}

#--- For debugging
sub _dump {
    my($self,$note) = @_;
    print STDERR "-- $note --\n";
    print STDERR "SIZE=$self->[SIZE]\n";
    print STDERR "USED=$self->[USED]\n";
    print STDERR "TIME=$self->[TIME]\n";
    while (my($k,$v) = each %{$self->[HASH]}) {
	print STDERR "'$k' => [$v->[POS]]\n";
    }
    for (my $i = 0; $i < $self->[USED]; $i ++) {
	my $irec = $self->[HEAP($i)];
	print STDERR "[$i] SEQ=$irec->[SEQ] POS=$irec->[POS] KEY=$irec->[KEY] VALUE=$irec->[VALUE]\n";
    }
}

1;
